//
//  Project   : besy
//  File      : server.cc
//  Author    : Ronald Kriemann / Christian Rost
//  Purpose   : class for a socketserver
//

extern "C"
{
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
}

#include <iostream.h>

#include "../misc/base64.hh"
#include "../misc/TRandom.hh"

#include "../misc/msqdef.hh"
#include "../misc/scanner.hh"

#include "server.hh"

// indicates if we have to finish
bool  bFinishWork;

// some defines for signing a coin
#define NOTHING          0
#define BEGIN_SIGNCOIN   1
#define RECV_COINDATA    2
#define REQUEST_BLINDFAC 3
#define RECV_KEYS        4
#define SIGN_COIN        5

// defines for rsa import
#define CLIENT  0
#define BANK    1
#define COIN    2
#define STORE   3

#define PRIVATE 0
#define PUBLIC  1

//
// signalhandler for SIGCHILD
//
// just gets the exitstatus and returns
//
void
sigHandler (int signr)
{
#ifdef DEBUG
	psignal( signr, "sigHandler (int) : recieved signal" );
#endif
	// handle signal
	switch (signr)
	{
        case SIGCHLD : // recieve childstatus
            waitpid( 0, (int*) NULL, WNOHANG );
            break;
        case SIGTERM : // finish process
            //	case SIGINT  :
            bFinishWork = TRUE;
            break;
        default      :
            break;
	}// switch
}

//
// constructor
//
Server::Server ( int portnumber )
{
	Socket.setPort(portnumber);
	bForked     = FALSE;
	msqid       = -1;
	bFinishWork = FALSE;
	init();
}

//
// destructor
//
Server::~Server ()
{
}

//
// create the serversocket / and install signalhandler
//
void 
Server::init ()
{
	struct sigaction rTERMAction;
  
	if (Socket.create() != SUCCESS)
	{
		cerr << "Server::init () : unable to create socket" << endl;
		exit( -1 );
	}// if

	// install all needed signalhandler
	signal(SIGCHLD, SIG_IGN);

	rTERMAction.sa_handler = (void (*) (int)) sigHandler;
	sigemptyset( &(rTERMAction.sa_mask) );
	sigaddset(   &(rTERMAction.sa_mask), SIGTERM );
	rTERMAction.sa_flags   = 0;
	sigaction( SIGTERM, &rTERMAction, NULL);
       
	bIDEA     = FALSE;
	bRSA      = TRUE;
	nRSAState = 0;

    rsaClient.setVerbosity( 1 );
    rsaBank.setVerbosity( 1 );
    rsaCoin.setVerbosity( 1 );

    signCoinState = NOTHING;

	// load all keys
	loadKey( BANK, 14, PRIVATE );
	loadKey( BANK, 14, PUBLIC );
    loadKey( COIN, 14, PRIVATE );
    loadKey( COIN, 14, PUBLIC );
}

//
// wait for clients
//
int 
Server::wait ()
{  
	int   sd;

	// wait for clients
	while (bFinishWork == FALSE)
    {            
		// check if clients have connected
		if ((sd = Socket.select()) >= 0)
		{
			// a client has connected and we have a communicationsocketdescriptor (nsd)  
			cout << "Server::wait () : client has connected" << endl;

// 			// fork another server
// 			pid_t pid = fork();

// 			// check if forked
// 			// then the parent starts commu with client and 
// 			// the child is listenning on the socketServer
// 			if (pid < 0)
// 			{
// 				perror( "Server::wait () : ERROR while forking" );
// 				exit( ERROR );
// 			}// if
// 			else if (pid == 0)
// 			{
            cout << "Server::wait () : coummnication-child has forked" << endl;

            bForked = TRUE; //I'm a child now
            // get the msqid for bank-communication
            key_t  key = MSQKEY;
            if ((msqid = msgget(key, MSQPERM)) < 0)
            {
                perror( "Server::wait () : ERROR while msgget" );
                return (ERROR);
            }// if
            // start commu with clients through nsd

            cout << "Server::run () : starting commu with client" << endl;

            return (run(sd));
// 			}// else
		}// if

    }// while
	return (SUCCESS);
}

//
// do your job
//
int 
Server::run (int nsd)
{
	MSGData  msgData;                  // Data for communication through msg_queue to bank
	bool     bEndCommu = FALSE;        // SocketCommunication alive?   
	TString  socketmsg;                // Message from socket 

	msgData.clear();
    
	while (bEndCommu == FALSE)       
    {
		// Waiting for a line\n from Client
		if (Socket.recvln(nsd,socketmsg) < 0) 
			// Client closes Connection
			bEndCommu = TRUE;      
	
		if (socketmsg.length() > 0)
		{	      	  	  
			// socketmsg enthaelt nun die erhaltene Nachricht
            // remove "\n"
			socketmsg.setLength(socketmsg.length()-1);

			// RSA-Login
			if (bRSA && (nRSAState < 2) && (socketmsg.length() > 0))
			{
                if (socketmsg.substr(0, 4) == "FAIL" )
                    nRSAState = NOTHING;
                else
                    rsaLogin( nsd, socketmsg, msgData );
			}// if
			else
			{
				if (bIDEA)
				{
					socketmsg = idea.decode(TBase64::decode(socketmsg));	      
                    cout << "Server::run() : length = " << socketmsg.length() << endl;
                    cout << "Server::run() :    msg = " << socketmsg << endl;
				}// if

                if (signCoinState != NOTHING)
                    signCoin( nsd, socketmsg );
                else
                {
                    // parse message
                    parseMsg( nsd, socketmsg, msgData, bEndCommu );
                }// else
			}// else
		}//if [Nachricht vorhanden]            
		else
		{
			bEndCommu = TRUE;
		}// else
    }// while  

	cout << "Server::run () : finished client communication" << endl;

	// close communication socket of this client
	Socket.close(nsd);

	return SUCCESS;
}

//  
// do parsing of messages
//
bool 
Server::parseMsg ( int  nsd, TString plainMsg, MSGData &msgData, bool &bEndCommu )
{
	TMessage msg;
	Scanner::TokenData data;
	bool     bSendMsg  = FALSE;
	bool     bAdd      = FALSE;
	  
	Scanner scanner( plainMsg );

	while ((scanner.nextToken(data) == SUCCESS) && (data.token != tokEOL))
	{	
        cout << "Data.token:" << data.token << endl;
		switch (data.token)
		{
            case tokKTNR :                               // KTNR		
                msgData.setKontoNr( data.nInt );
                cout << "Server::parseMsg (...) : set KTNR = " << data.nInt << endl;
                // Kontonummer wurde gesetzt
                sendln(nsd,"OK");
                break;
		  
            case tokNAME :                               // NAME
                cout << "Server::parseMsg () : name = " << data.szString << endl;
                msgData.setKundenname( data.szString );
                sendln( nsd, "OK");
                break;
		  
            case tokNUMMER :                              // Kundennummer
                cout << "Server::parseMsg () : Kundennummer = " << data.nInt << endl;
                msgData.setKundenNr( data.nInt );
                sendln(nsd,"OK");
                break;
		  
            case tokPASSWORT :                               // PASSWORT
                cout << "Server::parseMsg () : Passwort = " << data.szString << endl;
                msgData.setKundenPwd( data.szString );
                // LOGIN		  		     
                msgData.setOperation(LOGIN);				
                if ( msgData.hasKundenNr() && msgData.hasKundenname() )
                {				     				    
                    bSendMsg = TRUE;
                    msg.buildBankMsg(msgData);
                    //sendlnString( nsd, "OK" );
                }
                else 
                    // number is missing
                    sendln( nsd, "FAILED Missing Number or Name");
                break;		   			  
		  
            case tokONEPW :                               // OneTIME-PASSWORT
                cout << "Server::parseMsg () : Onetimepasw = " << data.szString << endl;
                msgData.setKundenPwd( data.szString );
                // ONELOGIN		  		     
                msgData.setOperation(ONELOGIN);
                if ( msgData.hasKundenNr() && msgData.hasKundenname())
                {
		      
                    msg.buildBankMsg(msgData);
                    bSendMsg = TRUE;
                    //sendln( nsd, "OK");
                } 
                else 
                    // number or Name is missing
                    sendln(nsd, "FAILED Missing Number or Name");
                break;
		  
            case tokZKTNR :                              // ZKTNR
                cout << "Server::parseMsg () : Zielkontonummer = " << data.nInt << endl;
                msgData.setZielKontoNr( data.nInt );
                sendln( nsd, "OK");
                break;
		  
            case tokZNAME :                              // ZNAME
                cout << "Server::parseMsg () : Zielname = " << data.szString << endl;
                msgData.setZielname( data.szString );
                sendln( nsd, "OK");
                break;
		  
		  
            case tokZNR :                              // ZNR
                cout << "Server::parseMsg () : Zielnummer = " << data.nInt << endl;
                msgData.setZielnummer( data.nInt );
                sendln( nsd, "OK");
                break;
		  
		  
            case tokBETRAG :                             // BETRAG
                cout << "Server::parseMsg () : Betrag = " << data.fFloat << endl;
                msgData.setBetrag( data.fFloat );
                sendln( nsd, "OK");
                break;
		  
            case tokZWECK :                              // ZWECK
                cout << "Server::parseMsg () : Zweck = " << data.szString << endl;
                msgData.setZweck( data.szString );
                // Operation ZWECK loesst die Operation Ueberweisen aus			
                if (     msgData.getOperation()==UEBERW &&
                         ((msgData.hasKundenNr() && msgData.hasKundenPwd())) &&
                         msgData.hasKontoNr()  && msgData.hasZielKontoNr() && 
                         msgData.hasZielname() && msgData.hasBetrag() && msgData.hasZweck() )				  
                    // Sind alle Daten vorhanden -> Dann los
                {	
			  
                    msg.buildBankMsg(msgData);
                    bSendMsg = TRUE;
                    //sendString( nsd, "OK Sending [Ueberw.] to Bank", bIDEA );
                    break;
                }// if
                else
                    // Zweck wurde gesetzt
                    sendln( nsd, "OK");				
                break;
		  
            case tokENDE :                               // ENDE
                cout << "Server::parseMsg () : closing connection to client" << endl;
                msgData.clear();
                msgData.setOperation(ENDE);
                msg.buildBankMsg(msgData);
                bSendMsg = TRUE;
                sendln(nsd,"OK Connection closed");
                // close connection			
                bEndCommu = TRUE;
                break;
		  
            case tokDot :
                cout << "Server::parseMsg () : dot received" << endl;
                //Falls Daten zum Rechtetransfer bekannt, ab zur Bank damit 
                if (  ((msgData.hasKundenNr() && msgData.hasKundenPwd())) && 
                      ( msgData.hasRechtAbheben() || msgData.hasRechtKontoauszug() )
                      && msgData.hasZielnummer()  && msgData.hasKontoNr() )
                {
                    msg.buildBankMsg(msgData);
                    bSendMsg = TRUE;				     
                } 
                else sendln( nsd, "FAILED Missing some data for this Operation");
                break;
		  
                // Change the permission			 			  
            case tokADD : bAdd=TRUE;
            case tokSUB : bAdd=(FALSE || bAdd);
                msgData.setOperation(RECHTE);
                if ((scanner.nextToken(data) == SUCCESS))
                {			     			     
                    switch (data.token)
                    {
                        case tokABHEBEN : // Recht zu ueberweisen
                            cout << "Server::parseMsg () :client wants to add/sub perm UEBERW"
                                 << endl;
                            // Das Recht Kodieren
                            if (bAdd) msgData.setRechtAbheben(1);
                            else msgData.setRechtAbheben(-1);
                            sendln( nsd, "OK");
                            break;
                        case tokKONTOAUSZUG :  // Recht fuer Kontoauszuege
                            cout << "Server::parseMsg () : client wants to add/sub perm KAUSZ"
                                 << endl;
                            // Das Recht Kodieren
                            if (bAdd) msgData.setRechtKontoauszug(1);
                            else msgData.setRechtKontoauszug(-1);
                            sendln( nsd, "OK");
                            break;
                        default :
                            nRSAState = signCoinState = NOTHING;
                            sendln( nsd,"FAILED Unknown Permission");
                            break;
                    }			      
                }//if
                else sendln(nsd,"FAILED Unexpected command");
                bAdd=FALSE;
                break; // end of permission
		  
            case tokOP :                                // OPERATION
                cout << "Server::parseMsg () : Operation requested" << endl;
                scanner.nextToken(data);

                switch (data.token)
                {
                    case tokOPKSTAND :                  // OP KSTAND
                        if ( msgData.hasKundenNr()   &&
                             msgData.hasKundenPwd() &&
                             msgData.hasKontoNr() )
                            // Sind alle Daten vorhanden -> Dann los
                        {
                            msgData.setOperation(KSTAND);
                            msg.buildBankMsg(msgData);
                            bSendMsg = TRUE;
                            //sendString( nsd, "OK sending message [KSTAND] to bank" );
                        }// if
                        else				
                            sendln(nsd, "FAILED Missing some data");		
                        break;
                        
                    case tokOPKAUSZ : 	                // OP AUSZ
                        msgData.setOperation( KAUSZ );
                        cout << msgData << endl;
                        if ( msgData.hasKundenNr()  &&
                             msgData.hasKundenPwd() &&
                             msgData.hasKontoNr() )
                            // Sind alle Daten vorhanden > Ab zur Bank damit
                        {
                            cout << "Server::parseMsg (...) : requested KAUSZ" << endl;
                            msgData.setOperation(KAUSZ);
                            msg.buildBankMsg(msgData);
                            bSendMsg = TRUE;							
                            //sendString( nsd, "OK sending message [KAUSZ] to bank" );
                        }// if
                        else
                        {
                            cout << "Server::parseMsg (...) : hasKundenNr  = "
                                 << msgData.hasKundenNr() << endl;
                            cout << "Server::parseMsg (...) : hasKundenPwd = "
                                 << msgData.hasKundenPwd() << endl;
                            cout << "Server::parseMsg (...) : hasKontoNr   = "
                                 << msgData.hasKontoNr() << endl;
                            sendln( nsd, "FAILED Missing some Data");
                        }// else
                        break;
                        
                    case tokOPUEBERW :                  // OP UEBERW
                        msgData.setOperation(UEBERW);
                        if ( msgData.hasKundenNr()    &&
                             msgData.hasKundenPwd()   && 
                             msgData.hasKontoNr()     &&
                             msgData.hasZielKontoNr() && 
                             msgData.hasZielname()    &&
                             msgData.hasBetrag()      &&
                             msgData.hasZweck() )
                        { // Alle Daten sind da -> Zur Bank					      
                            msg.buildBankMsg(msgData);
                            bSendMsg = TRUE;
                            //sendString( nsd, "OK sending message [UEBERW] to bank" );
                        }// if
                        else   					  
                            sendln( nsd, "FAILED Missing some Data");
                        
                        
                        break;
                        
                    case tokOPRECHTE :                  // OP RECHTE				
                        cout << "Server::parseMsg () : Change of Permission requested" 
                             << endl;
                        msgData.setOperation(RECHTE);
                        sendln(nsd, "OK");
                        break;							      			      
                        
                    case tokOPSignCoin :                  // OP SIGNCOIN
                        cout << "Server::parseMsg () : signing coin" << endl;
                        signCoinState = BEGIN_SIGNCOIN;
                        signCoin( nsd, data.szString );
                        break;
                        
                    default :                           // unknown OP
                        cout << "Server::parseMsg () : unknown operation : " 
                             << data.szString << endl;
                        nRSAState = signCoinState = NOTHING;
                        sendln(nsd,"FAILED Unknown Operation");
                        break;
                }// switch
                break;
                
            case tokNothing :
            default      :                               // unknown command
                // unknown command
                nRSAState = signCoinState = NOTHING;
                cout << "Server::parseMsg () : unknown command : " 
                     << plainMsg << endl ;                     
                sendln( nsd, "FAILED Unknown Command");
                break;
		}// switch
	}// while
	
	cout << "Server::parseMsg () : end of scanning msg, op_req = "<< bSendMsg << endl;
	
	// Operation requested ?
	if (bSendMsg)
	{					      		    		    
	    // delete old operation-data except logininformation and kontonr.
	    msgData.clear();
	    bSendMsg = FALSE;
        //send Message to Bank and send Result to clients
	    return sendln( nsd, sndMsg(msg) );
	} // end of Operation
	
	return SUCCESS;
}


//Send Message to Bank and return the resulting message
TString
Server::sndMsg ( TMessage &msg )
{  
	int                  size;
	bool                 bEndOfTransmission = FALSE;
	struct MessageBuffer mb;
    
	// for ident. we set the mb.type to our pid
	mb.msq_type = getpid();

	// send command, coded with our pid, to bank
    cout << "Server::sndMsg (TMessage) : length of msg = " << strlen(msg.getMsg()) << endl;
	strcpy(mb.msq_text,msg.getMsg());  

	// Message-Type == Receiver is 1  (==server)
	mb.msq_type=1;

    //
    // send msg to bank
    //
 	cout << "Server::sndMsg () : sending message to bank [id = " << msqid << "]:"
 		 << mb.msq_text << "[len = " << strlen(mb.msq_text) << "]" << endl;
	if (msgsnd( msqid, (struct msgbuf*) &mb, strlen(msg.getMsg()), 0) < 0)
    {
		perror ( "Server::sndMsg () : error in msgsnd" );
		return "Error while sending Message to bank";
    }// if

    //
	// wait for messages from bank with our pid and pass it to client
    //
// 	cout << "Server::sndMsg () : waiting for message from bank: " << endl;
	while ( !bEndOfTransmission &&
			(size = msgrcv( msqid, (struct msgbuf*) &mb, MSQSIZE, getpid(), 0)) > 0)
    {				  				  
		mb.msq_text[size] = 0;
    }// while

 	cout << "Server::sndMsg () : message form bank received: " << mb.msq_text << endl;
    
	return mb.msq_text;
}

//
// sends the line + \n to a socket
//
bool 
Server::sendln ( int nsd, const TString & string )
{
    TString  szMsg;

	if (bIDEA)
    {
        cout << "Server::sendln (...) : send = " << string << endl;
		szMsg = TBase64::encode(idea.encode(string));
    }// if
    else
        szMsg = string;
    
	return Socket.sendln( nsd, szMsg );
}

//
// RSA login
//
void
Server::rsaLogin (int nsd, const TString & socketmsg, MSGData &msgData )
{
	TString    decrypted, de1, de2;
	char       string[1024];
	TMessage   msg;  
	TString    recv;
	int        nounce1, nounce2;
	int        size;
	idea_key_t ideakey;

	nRSAState++;

	switch (nRSAState)
    {		  
        case 1 :
            //
            // recieving number and nounce
            //
            decrypted = rsaBank.decryptPrivate( TBase64::decode( socketmsg ) );
            decrypted <<= 4;
            decrypted >> rsaKNr;
            decrypted <<= 1;
            decrypted >> nounce1;

            if (loadKey( CLIENT, rsaKNr, PUBLIC))
            {
                nRSANounce = TRandom::randomInt();

                cout << "Server::rsaLogin (...) : got msg with kdnr = " << rsaKNr << endl
                     << "                         and       nounce1 = " << nounce1 << endl
                     << "                         sending   nounce2 = " << nRSANounce << endl;
            
                sprintf( string, "%d,%d", nounce1, nRSANounce);
                sendln( nsd, TBase64::encode( rsaClient.encryptPublic( string ) ) );
            }// if
            else 
            {
                cout << "Server::rsaLogin (...) : wrong kdnr = " << rsaKNr
                     << " was given" << endl;
            
                sendln( nsd, "FAIL : no key for given KDNR present" );
                nRSAState = NOTHING;
            }// else
            break;

        case 2 :
            //
            // receiving IDEA-Key
            //
            decrypted = rsaBank.decryptPrivate( TBase64::decode( socketmsg ) );
            cout << "Server::rsaLogin (...) :    length = " << decrypted.length() << endl;
            cout << "Server::rsaLogin (...) : decrypted = " << decrypted << endl;
            
            if (rsaClient.checkSign( decrypted ))
                rsaClient.removeSign( decrypted );
            else
            {
                cerr << "Server::rsaLogin (...) : msg is not signed correctly" << endl;
                sendln(nsd,"FAIL : msg is not signed correctly");
                nRSAState = NOTHING;
                return;
            }// else

            decrypted.copyTo( string, sizeof(string) );
            size = 0;
            // read until ','
            while ((string[size] != ',') && (size <= 160))
                size++; 

            nounce2 = -1;
            sscanf( string, "%d,", &nounce2);
		
            // checking nounce
            if (nounce2 == nRSANounce)
            {
                uchar  buffer[16];

                cout << "Server::rsaLogin (...) : got correct nounce2 = " << nounce2 << endl
                     << "                         extracting IDEA-key" << endl;
                memcpy(ideakey,&string[size+1],16);
                memcpy(buffer,&string[size+1],16);

                msgData.setKundenNr(rsaKNr);	
                msgData.setKundenname("RSA");
                msgData.setKundenPwd("RSA");
                sendln(nsd,"OK");

                cout << "Server::rsaLogin (...) : IDEAKEY = ";
                for (int i = 0; i < 16; i++)
                    cout << (uint) buffer[i] << ' ';
                cout << endl;

                idea.importKey(ideakey);
                idea.generateSubKeys();
                bIDEA=TRUE;
                bRSA=-1;
            }// if
            else
            {
                cout << "Server::rsaLogin (...) : wrong nounce2 = " << nounce2
                     << " was given" << endl;
                sendln(nsd,"FAIL : rcvd nounce is different from sent");
                nRSAState = NOTHING;
            }// else   
            break;

        default:
            nRSAState = NOTHING;
            break;
    }// switch
}	    

bool
Server::loadKey ( int owner, int number, int type )
{
    FILE  *file;
    char   buffer[16384];
    int    n;
    char   key[16384];

    switch (type)
    {
        case PRIVATE : file = fopen( "../keys/private_keys", "r" ); break;
        case PUBLIC  : file = fopen( "../keys/public_keys", "r" ); break;
        default        : return FALSE;
    }// switch

    if (file == NULL)
    {
        cerr << "Server::loadKey (...) : could not open file" << endl;
        return FALSE;
    }// if

    while (!feof( file ))
    {
        fgets( buffer, sizeof(buffer), file );
        
        memset( key, 0, sizeof(key) );
        n = -1;
        
        switch (owner)
        {
            case CLIENT : sscanf( buffer, "K_BESY%d,%s", &n, (char*) &key ); break;
            case BANK   : sscanf( buffer, "B_BESY%d,%s", &n, (char*) &key ); break;
            case COIN   : sscanf( buffer, "C_BESY%d,%s", &n, (char*) &key ); break;
            case STORE  : sscanf( buffer, "S_BESY%d,%s", &n, (char*) &key ); break;
            default       :
                fclose( file );
                return FALSE;
        }// switch

        if (n == number)
        {
            if (type == PRIVATE)
            {
                switch (owner)
                {
                    case CLIENT : rsaClient.setPrivateKey( TBase64::decode( key ) ); break;
                    case BANK   : rsaBank.setPrivateKey( TBase64::decode( key ) ); break;
                    case COIN   : rsaCoin.setPrivateKey( TBase64::decode( key ) ); break;
                    case STORE  : rsaStore.setPrivateKey( TBase64::decode( key ) ); break;
                }// switch
            }// if
            else // type == PUBLIC
            {
                switch (owner)
                {
                    case CLIENT : rsaClient.setModulus( TBase64::decode( key ) ); break;
                    case BANK   : rsaBank.setModulus( TBase64::decode( key ) ); break;
                    case COIN   : rsaCoin.setModulus( TBase64::decode( key ) ); break;
                    case STORE  : rsaStore.setModulus( TBase64::decode( key ) ); break;
                }// switch
            }// else

            fclose( file );
            
            return TRUE;
        }// if
    }// while

    fclose( file );

    return FALSE;
}

//
// take coin, requests blind-factors from client,
// sends coin and blindfaktors to bank and signed coin back to client
//
void
Server::signCoin ( int clientSocket, const TString &socketMsg )
{
    int        i;
    static int        k;
    char       buf[1024];
    static TString    coin;
    TString    string, signedCoin;
    TMD5       md5;
    static int        coinPos;
    static TBitArray  arrKeys[NO_OF_COIN];
    TIdea      coinIdea;

    TBigInteger<RSA_BASE>  bnBlind, bnHash;

    // Data for communication through msg_queue
    MSGData  msgData;

    // Formated Message of the above data
    TMessage msg;

    
    string = socketMsg;
    
    if (signCoinState == BEGIN_SIGNCOIN)
    {
        signCoinState = RECV_COINDATA;
        coinPos = 0;
    }// if
    else if (signCoinState == RECV_COINDATA)
    {
        if ((coinPos % 2) == 0)
        {
            cout << "Server::signCoin (...) : recv coin ( " << coinPos/2 << ")" << endl;
            arrCoinStr[coinPos/2] = TBase64::decode( string );
        }// if
        else
        {
            cout << "Server::signCoin (...) : recv hash ( " << coinPos/2 << ")" << endl;
            arrCoinMD5[coinPos/2] = TBase64::decode( string );
        }// else
        
        coinPos++;

        if (coinPos == (2*NO_OF_COIN))
            signCoinState = REQUEST_BLINDFAC;
    }// else
    else if (signCoinState == RECV_KEYS)
    {
        cout << "Server::signCoin (...) : recv key ( " << coinPos << ")" << endl;

        if (coinPos == k)
            coinPos++;
        
        if (coinPos >= NO_OF_COIN)
            signCoinState++;
        
        if (coinPos < NO_OF_COIN)
        {
            arrKeys[coinPos] = TBase64::decode( string );
            coinPos++;

            if (coinPos >= NO_OF_COIN)
                signCoinState++;
        }// if
    }// if
     
    cout << "signCoinState:" << signCoinState;
    switch (signCoinState)
    {
        case BEGIN_SIGNCOIN :
        case RECV_COINDATA  :
        case RECV_KEYS      :
            sendln( clientSocket, "OK" );
            return;
            
        case REQUEST_BLINDFAC :
            //  
            // build a random k between 0 and _NO_OF_BLINDFAC_ and
            // request all blindfactors except k
            //  

            k = TRandom::randomInt( 1, NO_OF_COIN-1 );
	    	    

            cout << "Server::signCoin (...) : requesting all blinds but " << k << endl;

            sprintf( buf, "OK %d", k );
            sendln( clientSocket, buf );

            coinPos = 0;
            
            break;

        case SIGN_COIN :
            //
            // decrypt md5-values of coins
            //
            for (i = 0; i < NO_OF_COIN; i++)
            {
                if (i != k)
                {
                    bnBlind = TBigInteger<RSA_BASE>( arrKeys[i].toString() );
                    bnBlind.powerMod( rsaBank.getPublicKey(), rsaBank.getModulus() );

                    bnHash  = TBigInteger<RSA_BASE>( arrCoinMD5[i] );
                    bnHash *= bnBlind.invMod( rsaBank.getModulus() );
                    bnHash %= rsaBank.getModulus();

                    arrCoinMD5[i] = bnHash.toString();
                }// if
            }// for
            
            //
            // check all coins
            //
            for (i = 0; i < NO_OF_COIN; i++)
            {
                if (i != k)
                {
                    coinIdea.generateSubKeys( arrKeys[i].toString() );
                    arrCoinStr[i] = coinIdea.decode( arrCoinStr[i] );

                    if (md5.crypt( arrCoinStr[i] ) != arrCoinMD5[i])
                    {
                        sendln( clientSocket, "FAIL : hash is not correct" );
                        signCoinState = NOTHING;
                        return;
                    }// if

                    if (!checkCoin( arrCoinStr[i] ))
                    {
                        sendln( clientSocket, "FAIL : coin is not correct" );
                        signCoinState = NOTHING;
                        return;
                    }// if
                }// if
            }// for

            //
            // sign hash and return to client
            //
            string  = "OK ";
            string += TBase64::encode( rsaBank.encryptPrivate( arrCoinMD5[k] ) );
            sendln( clientSocket, string );
            
            signCoinState = NOTHING;
            break;

        default:
            signCoinState = NOTHING;
            break;
    }// switch

    if (signCoinState != NOTHING)
        signCoinState++;
}

bool
Server::checkCoin ( const TString & string )
{
    uint     i, j;
    TMD5     md5;
    int      charSize = TString::charSize();
    
    TString  szCoin;
    TString  szBlindFac;
    TString  signedCoin;
    TString  temp( string ), subStr;

    TString   arrBlindMD5[NO_OF_ID][2];
    TString   arrR[NO_OF_ID][2];
    TBitArray arrID[NO_OF_ID][2];
    TBitArray refID, id;

    int      k = -1;
    TFloat   value;
    long     serialNo;

    ////////////////////////////////////////////////////////////
    //
    // check coin for correct format and blindfac
    //

    temp = string;

    // the value
    subStr = temp.strtok( ',' );
    subStr >> value;
    cout << "Server::signCoin (...) :    value = " << value << endl;

    if ((value != 1 ) && (value != 2 ) && (value != 5) &&
        (value != 10) && (value != 20) && (value != 50))
    {
        cerr << "Server::signCoin (...) : value = " << value
             << " is not in {1,2,5,10,20,50}" << endl;

        return FALSE;
    }// if

    // the bankID
    subStr = temp.strtok( ',' );
    cout << "Server::signCoin (...) :   bankID = " << subStr << endl;

    if (subStr != "B_BESY14")
    {
        cerr << "Server::signCoin (...) : not correct bankID" << endl;
        return FALSE;
    }// if

    // the serial no
    subStr = temp.strtok( ',' );
    subStr >> serialNo;
    cout << "Server::signCoin (...) : serialNo = " << serialNo << endl;

    if (hasSerialNo( serialNo ))
    {
        cerr << "Server::signCoin (...) : already have a coin with this serialNo = "
             << serialNo << endl;
        return FALSE;
    }// if

    return TRUE;
    
    // the md5'ed blind factors
    for (i = 0; i < NO_OF_ID; i++)
    {
        temp.strtok( '(' );

        subStr  = TBase64::decode( temp.strtok( ',' ) );
        subStr += TBase64::decode( temp.strtok( ')' ) );

        arrBlindMD5[i][0] = subStr.substr( 0, 128 / charSize );
        arrBlindMD5[i][1] = subStr.substr( (128 + BIT_COMMIT_SIZE) / charSize
                                           , 128 / charSize );

        arrR[i][0] = subStr.substr( 128 / charSize, BIT_COMMIT_SIZE / charSize );
        arrR[i][1] = subStr.substr( (256 + BIT_COMMIT_SIZE) / charSize
                                    , BIT_COMMIT_SIZE / charSize );

//         cout << "Server::signCoin (...) : md5(blind[" << i << ",0]) = "
//              << TBase64::encode(arrBlindMD5[i][0]) << endl;
//         cout << "Server::signCoin (...) :         R[" << i << ",0]) = "
//              << TBase64::encode(arrR[i][0]) << endl;
//         cout << "Server::signCoin (...) : md5(blind[" << i << ",1]) = "
//              << TBase64::encode(arrBlindMD5[i][1]) << endl;
//         cout << "Server::signCoin (...) :         R[" << i << ",0]) = "
//              << TBase64::encode(arrR[i][1]) << endl;

        if ((arrBlindMD5[i][0].length() != 16) || (arrBlindMD5[i][1].length() != 16) ||
            (arrR[i][0].length() != 32) || (arrR[i][1].length() != 32))
        {
            cerr << "Server::signCoin (...) : md5 and blindfac do not have correct length" << endl;
            return "FAIL : md5 and blindfac do not have correct length";
        }// if
        
        temp.strtok( ',' );
    }// for

    //
    // get and check blindfactors
    //

    temp = szBlindFac;
    
    for (i = 0; i < NO_OF_ID; i++)
    {
        if (i != uint(k))
        {
            arrID[i][0] = TBase64::decode( temp.strtok( ',' ) );
            arrID[i][1] = TBase64::decode( temp.strtok( ',' ) );

//             cout << "Server::signCoin (...) : ID[" << i << ",0] = "
//                  << TBase64::encode( arrID[i][0].toString() ) << endl;
//             cout << "Server::signCoin (...) : ID[" << i << ",1] = "
//                  << TBase64::encode( arrID[i][1].toString() ) << endl;
        }// if
    }// for

    if (k != 0)
    {
        refID  = arrID[0][0];
        refID ^= arrID[0][1];
    }// if
    else
    {
        refID  = arrID[1][0];
        refID ^= arrID[1][1];
    }// else
    
    for (i = 0; i < NO_OF_ID; i++)
    {
        if (i != uint(k))
        {
            id  = arrID[i][0];
            id ^= arrID[i][1];

            if (id != refID)
            {
                cerr << "Server::signCoin (...) : given ID's are not equal" << endl;
                return "FAIL : given ID's are not equal";
            }// if

            for (j = 0; j < 2; j++)
            {
                temp  = arrR[i][j];
                temp += arrID[i][j].toString();
                
//                 cout << "Server::signCoin (...) : md5(blind[" << i << "," << j << "]) = "
//                      << TBase64::encode(md5.crypt( temp )) << endl;
                
                if (md5.crypt( temp ) != arrBlindMD5[i][j])
                {
                    cerr << "Server::signCoin (...) : blindfactors are not correct" << endl;
                    return "FAIL : blindfactors are not correct";
                }// if
            }// for
        }// if
    }// for

    return "OK";
}

bool
Server::hasSerialNo ( int serialNo )
{
    MSGData  msgData;
    TMessage msg;
    TString  string;

    msgData.setOperation( CHECKSERIAL );
    msgData.setSerialNo( serialNo );

    msg.buildBankMsg( msgData );
    string = sndMsg( msg );

    if (string == "TRUE")
        return TRUE;
    else
        return FALSE;
}
