#ifndef __TBIGINTEGER_HH
#define __TBIGINTEGER_HH
//
//  Project   : besy
//  File      : TBigInteger.hh
//  Author    : Ronald Kriemann
//  Purpose   : class for big integers (bigger than long long long ...)
//  Date      : 01.12.1997
//

#include <math.h>

#include "defines.hh"
#include "TException.hh"
#include "dynarray.hh"
#include "TString.hh"

//
// the class uses the radix-b representation of an integer
// where b \in {2..255} is coded as the template parametre N
//
template <UINT4 N>
class TBigInteger : public TDynArray<UINT4>
{
protected:
    // some flags
    bool   bUnderFlow, bOverFlow, bDivByZero;
    
public:
    // constructors
    TBigInteger ();
    TBigInteger ( const TBigInteger<N> &n );
    TBigInteger ( const ulong &n );
    TBigInteger ( const TString & string );

    // destructor
    ~TBigInteger ();

    // get radix-b
    const UINT4 getRadix () const;

    //
    // access flags
    //
    const bool isUnderFlow () const;
    const bool isOverFlow () const;
    const bool isDivByZero () const;

    //
    // usual operators
    //

    // copy
    TBigInteger<N> & operator = ( const ulong &n );

    // add + sub
    TBigInteger<N> & operator += ( const TBigInteger<N> &n );
    
    TBigInteger<N> & operator -= ( const TBigInteger<N> &n );

    // multiplication + division + mod
    TBigInteger<N> & operator *= ( const TBigInteger<N> &n );
    
    TBigInteger<N> & operator /= ( const TBigInteger<N> &n );

    TBigInteger<N> & operator %= ( const TBigInteger<N> &n );

    // order
    const bool operator < ( const TBigInteger<N> &n ) const;
    const bool operator > ( const TBigInteger<N> &n ) const;
    const bool operator == ( const TBigInteger<N> &n ) const;
    const bool operator != ( const TBigInteger<N> &n ) const;
    const bool operator <= ( const TBigInteger<N> &n ) const;
    const bool operator >= ( const TBigInteger<N> &n ) const;

    //
    // usual math functions
    //

    // power
    TBigInteger<N> & power ( const TBigInteger<N> &n );

    // power and mod ( x^n mod m )
    TBigInteger<N> & powerMod ( const TBigInteger<N> &n,
                                const TBigInteger<N> &m );

    // return inverse with respect to modulo
    TBigInteger<N> invMod ( const TBigInteger<N> & n ) const;
    
    //
    // misc. methods
    //

    // investigate math. properties
    const bool isEven () const;
    const bool isOdd () const;
    const bool isPrime () const;

    // remove leading 0's
    void shorten ();

    // import/export from/to strings (string[0] is considered lsb)
    void fromString ( const TString & string );
    TString toString () const;

    // output
#ifdef EGCS
    friend ostream & operator << <> ( ostream &os, const TBigInteger<N> &n );
#else
    friend ostream & operator <<    ( ostream &os, const TBigInteger<N> &n );
#endif
};

//////////////////////////////////////////////////////
//
// definition
//

//
// constructor
//
template <UINT4 N>
TBigInteger<N>::TBigInteger ()
{
    bUnderFlow = bOverFlow = bDivByZero = FALSE;

    // init with usefull settings
    setGrowBy( 256 );

    setSize( 1 );
    (*this)[0] = 0;
}

template <UINT4 N>
TBigInteger<N>::TBigInteger ( const TBigInteger<N> &n )
{
    ::TBigInteger<N>();
    
    *this = n;
}

template <UINT4 N>
TBigInteger<N>::TBigInteger ( const ulong &n )
{
    ::TBigInteger<N>();

    *this = n;
}

template <UINT4 N>
TBigInteger<N>::TBigInteger ( const TString & string )
{
    fromString( string );
}

//
// destructor
//
template <UINT4 N>
TBigInteger<N>::~TBigInteger ()
{}

// get radix-b
template <UINT4 N>
const UINT4 
TBigInteger<N>::getRadix () const
{
    return N;
}

//
// access flags
//
template <UINT4 N>
const bool 
TBigInteger<N>::isUnderFlow () const
{
    return bUnderFlow;
}

template <UINT4 N>
const bool 
TBigInteger<N>::isOverFlow () const
{
    return bOverFlow;
}

template <UINT4 N>
const bool 
TBigInteger<N>::isDivByZero () const
{
    return bDivByZero;
}

//
// copy
//
template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::operator = ( const ulong &n )
{
    uint    i = 0;
    UINT4   r = n;

    if (n == 0)
    {
        setSize( 1 );
        (*this)[0] = 0;
    }// if
    else
    {
        while (r > 0)
        {
            setSize( i+1 );
            (*this)[i++] = r % N;
            r /= N;
        }// while
    }// else        

    return *this;
}

//
// add + sub
//
template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::operator += ( const TBigInteger<N> &n )
{
    uint     i;
    uint     size;
    UINT4    result, overflow = 0;
    UINT4   *pData1;
    UINT4   *pData2;

    if (n.getSize() > getSize())
    {
        // add n to this
        size = n.getSize();
        setSize( size );

        pData1 =   getData();
        pData2 = n.getData();

        for (i = 0; i < size; i++)
        {
            result   = *pData1 + *pData2 + overflow;

            if (result >= N)
            {
                overflow = 1;
                result  %= N;
            }// if
            else
                overflow = 0;

            *pData1++ = result;
            pData2++;
        }// for

        // take care about the overflow
        if (overflow > 0)
        {
            setSize( size+1 );
            (*this)[size] = overflow;
        }// if
    }// if
    else
    {
        // add n to this
        size = n.getSize();

        pData1 =   getData();
        pData2 = n.getData();

        for (i = 0; i < size; i++)
        {
            result = *pData1 + *pData2 + overflow;

            if (result >= N)
            {
                overflow = 1;
                result  %= N;
            }// if
            else
                overflow = 0;

            *pData1++ = result;
            pData2++;
        }// for

        // take care about the overflow
        size = getSize();

        while (overflow > 0)
        {
            if (i >= size)
            {
                setSize( size+1 );
                (*this)[i] = overflow;

                overflow   = 0;
            }// if
            else
            {
                result = *pData1 + overflow;

                if (result >= N)
                {
                    overflow = 1;
                    result  %= N;
                }// if
                else
                    overflow = 0;
                
                *pData1 = result;
            }// else

            pData1++;
            i++;
        }// if
    }// else

    return *this;
}

template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::operator -= ( const TBigInteger<N> &n )
{
    uint    i;
    uint    size1, size2;
    UINT4   result, underflow = 0;
    UINT4   a1, a2;
    UINT4  *pData1;
    UINT4  *pData2;

    bUnderFlow = FALSE;

    // sub n from this
    size1 =   getSize();
    size2 = n.getSize();

    pData1 =   getData();
    pData2 = n.getData();

    if (size1 < size2)
    {
        for (i = 0; i < size1; i++)
        {
            a1 = *pData1;
            a2 = *pData2 + underflow;

            if (a1 < a2)
            {
                result = N + a1 - a2;
                underflow = 1;
            }// if
            else
            {
                result = a1 - a2;
                underflow = 0;
            }// else
            
            *pData1++ = result;
            pData2++;
        }// for
    }// if
    else
    {
        for (i = 0; i < size2; i++)
        {
            a1 = *pData1;
            a2 = *pData2 + underflow;

            if (a1 < a2)
            {
                result = N + a1 - a2;
                underflow = 1;
            }// if
            else
            {
                result = a1 - a2;
                underflow = 0;
            }// else
            
            *pData1++ = result;
            pData2++;
        }// for

        for (i = size2; i < size1; i++)
        {
            if (*pData1 < underflow)
            {
                result = N + *pData1 - underflow;
                underflow = 1;
            }// if
            else
            {
                result = *pData1 - underflow;
                underflow = 0;
            }// else
            
            *pData1++ = result;
        }// for
    }// else

    shorten();

    bUnderFlow = underflow;
#if defined(EXCEPTIONS)
    if (bUnderFlow)
        throw TUnderFlow();
#endif

    return *this;
}

//
// multiplication + division + mod
//
template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::operator *= ( const TBigInteger<N> &n )
{
    uint            i, j;
    uint            size1, size2, size3;
    UINT4           a2;
    uint            result, overflow;
    TBigInteger<N>  temp;

    UINT4  *pData1;
    UINT4  *pData2;
    UINT4  *pData3;

    size1 =   getSize();
    size2 = n.getSize();
    size3 = size1 + size2;

    // init. temp
    temp.setSize( size3 );
    pData3 = temp.getData();
    for (i = 0; i < size3; i++)
        *pData3++ = 0;

    pData1 = getData();
    pData2 = n.getData();
    pData3 = temp.getData();

    for (i = 0; i < size2; i++)
    {
        a2 = pData2[i];
        overflow = 0;

        if (a2 != 0)
        {
            for (j = 0; j < size1; j++)
            {
                result    = pData1[j] * a2 + overflow + pData3[i+j];
                overflow  = result / N;
                result   %= N;

                pData3[i+j] = result;
            }// for

            if (overflow > 0) pData3[i+size1] = overflow;
        }// if
    }// for

    temp.shorten();

    return *this = temp;
}

template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::operator /= ( const TBigInteger<N> &n )
{
    UINT4            d, q_temp;
    ulong            size_m, size_n, size, i, j, i_q;
    TBigInteger<N>   v, q;
    UINT4            overFlow, underFlow, result;
    bool             bOneRun;

    register UINT4  *pData1;
    register UINT4  *pData2;
    UINT4  *pData3;

    shorten();
    size   = getSize();
    size_n = n.getSize();

    // determine real size of n
    i = size_n - 1;
    while (i > 0)
    {
        if (n[i--] == 0)
            size_n--;
        else
            i = 0;
    }// while

    // trivial checks
    if (size < size_n) 
        return *this = 0;

    // divide by zero
    bDivByZero = FALSE;
    if (n == 0)
    {
#if defined(EXCEPTIONS)
        throw TDivByZero();
#endif
        bDivByZero = TRUE;
        return *this;
    }// if

    // divide by 1
    if (n == 1)
        return *this;

    // and this < n
    if (*this < n)
    {
        return *this = 0;
    }// if

    v = n;

    d = N / (v[size_n-1] + 1);
    if (d > 1)
    {
        *this *= d;
        v *= d;

        size   = getSize();
        size_n = v.getSize();
    }// if
    
    size_m = size - size_n;

    q.setSize( size_m + 1 );
    pData3 = q.getData();
    for (i = 0; i <= size_m; i++)
        *pData3++ = 0;

    i_q = 0;

    j = size - 1;

    pData1 =   getData();
    pData2 = v.getData();
    pData3 = q.getData();

    if (size == size_n)
        bOneRun = TRUE;
    else
        bOneRun = FALSE;

    while ((j >= size_n) || bOneRun)
    {
        if (size == size_n)
            q_temp = pData1[j] / pData2[size_n-1];
        else
            q_temp = (pData1[j]*N + pData1[j-1]) / pData2[size_n-1];

        if (size_n > 1)
        {
            while ((pData2[size_n-2]*q_temp) > 
                   ((pData1[j]*N + pData1[j-1] - 
                     (q_temp*pData2[size_n-1]))*N + pData1[j-2]))
            {
                q_temp--;
            }// while
        }// if
        
        underFlow = 0;
        if (q_temp != 0)
        {
            overFlow = 0;
            if (size == size_n)
            {
                for (i = 0; i < size_n; i++)
                {
                    result    = q_temp * pData2[i] + overFlow;
                    overFlow  = result / N;
                    result   %= N;
                
                    if (pData1[j+1-size_n+i] < result)
                    {
                        pData1[j+1-size_n+i] += N;
                        overFlow++;
                    }// if
                
                    pData1[j+1-size_n+i] -= result;
                }// for
            
                if (overFlow > 0)
                {
                    cout << "to be done" << endl;
                }// if
            }// if
            else
            {
                for (i = 0; i < size_n; i++)
                {
                    result    = q_temp * pData2[i] + overFlow;
                    overFlow  = result / N;
                    result   %= N;
                
                    if (pData1[j-size_n+i] < result)
                    {
                        pData1[j-size_n+i] += N;
                        overFlow++;
                    }// if
                
                    pData1[j-size_n+i] -= result;
                }// for
            
                if (overFlow > 0)
                {
                    if (pData1[j] < overFlow)
                    {
                        underFlow    = 1;
                        pData1[j] += N;
                    }// if
                
                    pData1[j] -= overFlow;
                }// if
            }// else
        }// if
        
        if (underFlow != 0)
        {
            q_temp--;

            // adjust the value by adding v
            overFlow = 0;
            for (i = 0; i < size_n; i++)
            {
                result = pData2[i] + pData1[j-size_n+i] + overFlow;
                
                if (result >= N)
                {
                    overFlow = 1;
                    result  %= N;
                }// if
                else
                    overFlow = 0;
                
                pData1[j-size_n+i] = result;
            }// for
            if (overFlow > 0)
            {
                result  = pData1[j] + overFlow;
                result %= N;

                pData1[j] = result;
            }// if
        }// if

        pData3[i_q] += q_temp;

        //        cout << "u = " << u << endl;
        //        cout << "q = " << q << endl;

        if ((pData1[j] == 0) || (q_temp == 0))
        {
            if (j > 0) j--;
            i_q++;
            bOneRun = FALSE;
        }// if
        else
            bOneRun = TRUE;
    }// while

    // copy q to this and check if all q's are  < N
    size = i_q;
    setSize( size );
    pData1 = getData();
    overFlow = 0;
    for (i = 0; i < size; i++)
    {
        result = pData3[size-i-1] + overFlow;
        if (result >= N)
        {
            overFlow  = result / N;
            result   %= N;
        }// if

        pData1[i] = result;
    }// if
    if (overFlow > 0)
    {
        setSize( size+1 );
        (*this)[size] = overFlow;
    }// if

    shorten();

    return *this;
}

template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::operator %= ( const TBigInteger<N> &n )
{
    UINT4            d, q_temp;
    ulong            size_n, size_m, size, i, j, i_q;
    TBigInteger<N>   v, q;
    UINT4            overFlow, underFlow, result;
    bool             bOneRun;

    register UINT4  *pData1;
    register UINT4  *pData2;

    shorten();
    size   =   getSize();
    size_n = n.getSize();

    // determine real size of n
    i = size_n - 1;
    while (i > 0)
    {
        if (n[i--] == 0)
            size_n--;
        else
            i = 0;
    }// while

    // trivial checks
    if (size < size_n) return *this;

    // mod zero
    bDivByZero = FALSE;
    if (n == 0)
    {
#if defined(EXCEPTIONS)
        throw TDivByZero();
#endif
        bDivByZero = TRUE;
        return *this = 0;
    }// if

    // mod 1
    if (n == 1) return *this = 0;

    // and this < n
    if (*this < n) return *this;

    v = n;

    d = N / (v[size_n-1] + 1);
    if (d > 1)
    {
        *this *= d;
        v *= d;

        size   =   getSize();
        size_n = v.getSize();
    }// if
    
    size_m = size - size_n;

    j = size - 1;

    pData1 =   getData();
    pData2 = v.getData();
    i_q = 0;

    if (size == size_n)
        bOneRun = TRUE;
    else
        bOneRun = FALSE;

    while ((j >= size_n) || bOneRun)
    {
        if (size == size_n)
            q_temp = pData1[j] / pData2[size_n-1];
        else
            q_temp = (pData1[j]*N + pData1[j-1]) / pData2[size_n-1];

        if (size_n > 1)
        {
            while ((pData2[size_n-2]*q_temp) > 
                   ((pData1[j]*N + pData1[j-1] - 
                     (q_temp*pData2[size_n-1]))*N + pData1[j-2]))
            {
                q_temp--;
            }// while
        }// if
        
        underFlow = 0;
        if (q_temp != 0)
        {
            overFlow = 0;

            if (size == size_n)
            {
                for (i = 0; i < size_n; i++)
                {
                    result    = q_temp * pData2[i] + overFlow;
                    overFlow  = result / N;
                    result   %= N;
                
                    if (pData1[j+1-size_n+i] < result)
                    {
                        pData1[j+1-size_n+i] += N;
                        overFlow++;
                    }// if
                
                    pData1[j+1-size_n+i] -= result;
                }// for
            
                if (overFlow > 0)
                {
                    cout << "to be done" << endl;
                }// if
            }// if
            else
            {
                for (i = 0; i < size_n; i++)
                {
                    result    = q_temp * pData2[i] + overFlow;
                    overFlow  = result / N;
                    result   %= N;
                
                    if (pData1[j-size_n+i] < result)
                    {
                        pData1[j-size_n+i] += N;
                        overFlow++;
                    }// if
                
                    pData1[j-size_n+i] -= result;
                }// for
            
                if (overFlow > 0)
                {
                    if (pData1[j] < overFlow)
                    {
                        underFlow    = 1;
                        pData1[j] += N;
                    }// if
                
                    pData1[j] -= overFlow;
                }// if
            }// else
        }// if
        
        if (underFlow != 0)
        {
            q_temp--;

            // adjust the value by adding v
            overFlow = 0;
            for (i = 0; i < size_n; i++)
            {
                result = pData2[i] + pData1[j-size_n+i] + overFlow;
                
                if (result >= N)
                {
                    overFlow = 1;
                    result  %= N;
                }// if
                else
                    overFlow = 0;
                
                pData1[j-size_n+i] = result;
            }// for
            if (overFlow > 0)
            {
                result  = pData1[j] + overFlow;
                result %= N;

                pData1[j] = result;
            }// if
        }// if

        if ((pData1[j] == 0) || (q_temp == 0))
        {
            if (j > 0) j--;
            i_q++;
            bOneRun = FALSE;
        }// if
        else
            bOneRun = TRUE;
    }// while

    setSize( size_n );
    shorten();

    if (d > 1)
    {
        pData1 = getData();
        j      = getSize()-1;

        if (j == 0)
            pData1[0] /= d;
        else
        {
            size_m = size_n - 1;
            q.setSize( size_m + 1 );
            pData2 = q.getData();
            for (i = 0; i <= size_m; i++)
                *pData2++ = 0;

            pData2 = q.getData();
            i_q = 0;

            if (size == 1)
                bOneRun = TRUE;
            else
                bOneRun = FALSE;
            
            while ((j >= 1) || bOneRun)
            {
                bOneRun = FALSE;

                q_temp = (pData1[j]*N + pData1[j-1]) / d;
            
                underFlow = 0;
                if (q_temp != 0)
                {
                    result    = q_temp * d;
                    overFlow  = result / N;
                    result   %= N;
                
                    if (pData1[j-1] < result)
                    {
                        pData1[j-1] += N;
                        overFlow++;
                    }// if
                
                    pData1[j-1] -= result;
            
                    if (overFlow > 0)
                    {
                        if (pData1[j] < overFlow)
                        {
                            underFlow  = 1;
                            pData1[j] += N;
                        }// if
                
                        pData1[j] -= overFlow;
                    }// if
                }// if

                pData2[i_q] += q_temp;

                if (pData1[j] == 0) 
                {
                    j--;
                    i_q++;
                }// if
            }// while
            
            size = i_q;
            setSize( size );
            overFlow = 0;
            for (i = 0; i < size; i++)
            {
                result = pData2[size-i-1] + overFlow;
                if (result >= N)
                {
                    overFlow  = result / N;
                    result   %= N;
                }// if
                
                pData1[i] = result;
            }// if
            if (overFlow > 0)
            {
                setSize( size+1 );
                (*this)[size] = overFlow;
            }// if
        }// else
    }// if

    shorten();

    return *this;
}

//
// order
//
template <UINT4 N>
const bool 
TBigInteger<N>::operator < ( const TBigInteger<N> &n ) const
{
    int     i;
    int     size1, size2;
    UINT4   a1, a2;
    UINT4  *pData1;
    UINT4  *pData2;

    size1 =   getSize();
    size2 = n.getSize();

    pData1 =   getData();
    pData2 = n.getData();

    if (size2 < size1)
    {
        for ( i = size1-1; i >= size2; i-- )
            if (pData1[i] > 0)
                return FALSE;

        for ( i = size2-1; i >= 0; i-- )
        {
            a1 = pData1[i]; a2 = pData2[i];
            if (a1 > a2)
                return FALSE;
            else if (a1 < a2)
                return TRUE;
        }// for
    }// if
    else
    {
        for ( i = size2-1; i >= size1; i-- )
            if (pData2[i] > 0)
                return TRUE;

        for ( i = size1-1; i >= 0; i-- )
        {
            a1 = pData1[i]; a2 = pData2[i];
            if (a1 > a2)
                return FALSE;
            else if (a1 < a2)
                return TRUE;
        }// if
    }// else

    return FALSE;
}

template <UINT4 N>
const bool 
TBigInteger<N>::operator > ( const TBigInteger<N> &n ) const
{
    return n < *this;
}

template <UINT4 N>
const bool 
TBigInteger<N>::operator == ( const TBigInteger<N> &n ) const
{
    int   i;
    int   size1, size2;
    
    UINT4  *pData1;
    UINT4  *pData2;

    size1 =   getSize();
    size2 = n.getSize();

    pData1 =   getData();
    pData2 = n.getData();

    if (size2 < size1)
    {
        for ( i = 0; i < size2; i++ )
        {
            if (*pData1++ != *pData2++)
                return FALSE;
        }// for

        for ( i = size2; i < size1; i++ )
        {
            if (*pData1++ != 0)
                return FALSE;
        }// for
    }// if
    else
    {
        for ( i = 0; i < size1; i++ )
        {
            if (*pData1++ != *pData2++)
                return FALSE;
        }// for

        for ( i = size1; i < size2; i++ )
        {
            if (*pData2++ != 0)
                return FALSE;
        }// for
    }// else

    return TRUE;
}

template <UINT4 N>
const bool 
TBigInteger<N>::operator != ( const TBigInteger<N> &n ) const
{
    return !(*this == n);
}

template <UINT4 N>
const bool 
TBigInteger<N>::operator <= ( const TBigInteger<N> &n ) const
{
    int     i;
    int     size1, size2;
    UINT4   a1, a2;
    UINT4  *pData1;
    UINT4  *pData2;

    pData1 =   getData();
    pData2 = n.getData();

    size1 =   getSize();
    size2 = n.getSize();

    if (size2 < size1)
    {
        for ( i = size1-1; i >= size2; i-- )
            if (pData1[i] > 0)
                return FALSE;

        for ( i = size2-1; i >= 0; i-- )
        {
            a1 = pData1[i]; 
            a2 = pData2[i];

            if (a1 > a2)
                return FALSE;
            else if (a1 < a2)
                return TRUE;
        }// for
    }// if
    else
    {
        for ( i = size2-1; i >= size1; i-- )
            if (pData2[i] > 0)
                return TRUE;

        for ( i = size1-1; i >= 0; i-- )
        {
            a1 = pData1[i]; 
            a2 = pData2[i];

            if (a1 > a2)
                return FALSE;
            else if (a1 < a2)
                return TRUE;
        }// if
    }// else

    return TRUE;
}

template <UINT4 N>
const bool 
TBigInteger<N>::operator >= ( const TBigInteger<N> &n ) const
{
    return n <= *this;
}

// power
template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::power ( const TBigInteger<N> &n )
{
    int     i;
    UINT4   j;
    ulong   size = n.getSize();
    UINT4   a;

    TBigInteger<N>  s( 1 ), temp;

    if (N == 2)
    {
        for (i = size-1; i > 0; i--)
        {
            if (n[i] != 0) s *= *this;
            s *= s;
        }// for
        if (n[0] != 0) s *= *this;
    }// if
    else
    {
        for (i = size-1; i > 0; i--)
        {
            a = n[i];
            if (a != 0)
            {
                temp = *this;
                for (j = 1; j < a; j++)
                    temp *= *this;
                s *= temp;
            }// if

            temp = s;
            for (j = 1; j < N; j++)
                s *= temp;
        }// for

        a = n[0];
        if (a != 0)
        {
            temp = *this;
            for (j = 1; j < a; j++)
                temp *= *this;
            s *= temp;
        }// if
    }// else

    return *this = s;
}

//
// power and mod ( x^n mod m )
//
template <UINT4 N>
TBigInteger<N> & 
TBigInteger<N>::powerMod ( const TBigInteger<N> &n,
                           const TBigInteger<N> &m )
{
    char            arrRotate[] = "-/|\\-/|\\";
    int             i;
    ulong           size = n.getSize();
    UINT4           a, p, shift, base_shift;
    UINT4          *pData;
    TBigInteger<N>  s( 1 ), temp, k;

    pData = n.getData();

    if (N == 2)
    {
        for (i = size-1; i > 0; i--)
        {
            if (pData[i] != 0)
            {
                s *= *this;
                s %= m;
            }// if

            s *= s;
            s %= m;
        }// for
        if (pData[0] != 0)
        {
            s *= *this;
            s %= m;
        }// if
    }// if
    else
    {
        // determine real size of n
        i = size - 1;
        while (i > 0)
        {
            if (n[i--] == 0)
                size--;
            else
                i = 0;
        }// while

        if (size == 1)
        {
            k  = *this;
            k %= m;

            a = pData[0];
            if (a != 0)
            {
                temp = 1;
                
                p = 1;
                shift = ((uint)(log(a) / M_LN2));
                if (shift > 0) p <<= shift;
                
                while (p > 0)
                {
                    if ((a & p) != 0) 
                    {
                        s *= k;
                        s %= m;
                    }// if
                    
                    p >>= 1;

                    if (p > 0)
                    {
                        s *= s;
                        s %= m;
                    }// if
                }// for

                return *this = s;
            }// if
        }// if
        else
        {
            base_shift = ((uint)(log(N) / M_LN2));
            for (i = size-1; i > 0; i--)
            {
                cout << arrRotate[i % 8] << flush;

                a = pData[i];
                if (a != 0)
                {
                    temp = *this;
                    temp.powerMod( a, m );
                    s *= temp;
                    s %= m;
                }// if
            
                temp = 1;
                p = 1;
                if (base_shift > 0) p <<= base_shift;
                while (p > 0)
                {
                    if ((N & p) != 0) 
                    {
                        temp *= s;
                        temp %= m;
                    }// if
                    
                    p >>= 1;
                    
                    if (p > 0)
                    {
                        temp *= temp;
                        temp %= m;
                    }// if
                }// while
                
                s = temp;

                cout << char(8) << flush;
            }// for

            a = pData[0];
            if (a != 0)
            {
                temp = *this;
                temp.powerMod( a, m );
                s *= temp;
                s %= m;
            }// if
        }// else
    }// else

    return *this = s;
}

//
// return inverse with respect to modulo
//
template <UINT4 N>
TBigInteger<N>
TBigInteger<N>::invMod ( const TBigInteger<N> & n ) const
{
    bool            bIsB1Neg, bIsB2Neg, bOldB2;
    TBigInteger<N>  r, q, n1, n2, b1, b2, t, temp, inv;

    //
    // building the mult. inv. in Z_n
    // (with euclidean algo.)
    //
        
    n1 = n;
    n2 = *this;
    b2 = 1;
    b1 = 0;
    
    bIsB1Neg = bIsB2Neg = FALSE;
        
    while (TRUE)
    {
        // r = n1 % n2;
        r  = n1;
        r %= n2;
            
        // q = (n1 - r) / n2
        q  = n1;
        q -= r;
        q /= n2;
            
        if (r == 0)
        {
            if (bIsB2Neg) 
            {
                inv  = n;
                inv -= b2;
            }// if
            else
                inv = b2;
                
            break;
        }// if
            
        n1 = n2;
        n2 = r;
        t  = b2;
            
        bOldB2 = bIsB2Neg;
        temp  = b2;
        temp *= q;
            
        if (!bIsB1Neg && !bIsB2Neg)
        {
            if (b1 < temp)
            {
                bIsB2Neg = TRUE;
                b2  = temp;
                b2 -= b1;
            }// if
            else
            {
                b2  = b1;
                b2 -= temp;
            }// else
        }
        else if (bIsB1Neg && bIsB2Neg)
        {
            if (b1 < temp) bIsB2Neg = FALSE;
            b2  = temp;
            b2 -= b1;
        }// else
        else if (!bIsB1Neg && bIsB2Neg)
        {
            bIsB2Neg = FALSE;
            b2  = temp;
            b2 += b1;
        }// else
        else if (bIsB1Neg && !bIsB2Neg)
        {
            bIsB2Neg = TRUE;
            b2  = temp;
            b2 += b1;
        }// else
            
        bIsB1Neg = bOldB2;
        b1 = t;
    }// while

    return inv;
}
    
//
// investigate math. properties
//
template <UINT4 N>
const bool 
TBigInteger<N>::isEven () const
{
    if (((*this)[0] % 2) == 0)
        return TRUE;
    else
        return FALSE;
}

template <UINT4 N>
const bool 
TBigInteger<N>::isOdd () const
{
    if (((*this)[0] % 2) == 1)
        return TRUE;
    else
        return FALSE;
}

// array with primes
extern uint arrPrimes[];

template <UINT4 N>
const bool 
TBigInteger<N>::isPrime () const
{
    uint         i;
    uint         noOfPrimes = 167;
    TBigInteger  a, p_1, temp;

    // trivial cases
    if ((*this == 0) || (*this == 1) || (isEven() && (*this > 2)))
        return FALSE;

    //
    // now check if we can divide with some usual primes
    //
    for (i = 0; i < noOfPrimes; i++)
    {
        temp = *this;
        temp %= arrPrimes[i];
        if (temp == 0) return FALSE;
    }// for
    
    //
    // do primetest (Fermat; only for 2 because probability is < 10^-44)
    //
    a    = 2;
    p_1  = *this;
    p_1 -= 1;
    //    p_1 /= 2;  // this is Euler-test with Jacobi(2,this) == 2^((this-1)/2)

    a.powerMod( p_1, *this );

    if (a == 1)
        return TRUE;
    else
        return FALSE;
}

//
// remove leading 0's
//
template <UINT4 N>
void 
TBigInteger<N>::shorten ()
{
    ulong   size  = getSize();
    UINT4  *pData = getData();

    if (size == 0)
    {
        setSize( 1 );
        (*this)[0] = 0;
    }// if
    else if (size == 1) 
        return;
    else
    {
        size--;

        while ((size > 0) && (pData[size] == 0))
            size--;

        setSize( size+1 );
    }// else
}

//
// import/export from/to strings
//
template <UINT4 N>
void
TBigInteger<N>::fromString ( const TString & string )
{
    int             i;
    uint            len = string.length();
    UINT4           b;
    TBigInteger<N>  base;

    *this = 0;

    if (string.length() == 0)
        return;
    
    b   = 1;
    b <<= string.charSize();

    base = b;
    
    for (i = len-1; i > 0; i--)
    {
        *this += int(string[i]);
        *this *= base;
    }// for
    *this += int(string[0]);
}

template <UINT4 N>
TString
TBigInteger<N>::toString () const
{
    int             i, len;
    UINT4           b;
    TBigInteger<N>  temp( *this ), r, base;
    TString         string;
    uchar           coeff;

    b   = 1;
    b <<= string.charSize();

    base = b;

    string = "";
    
    while (temp != 0)
    {
        r  = temp;
        r %= base;

        temp /= base;

        len   = r.getSize();
        coeff = 0;

        if (len > 0)
        {
            for (i = len-1; i > 0; i--)
            {
                coeff += uchar( r[i] );
                coeff *= N;
            }// for
            coeff += uchar( r[0] );
        }// if
                 
        string += coeff;
    }// while

    return string;
}

//
// output
//
template <UINT4 N>
ostream & operator << ( ostream &os, const TBigInteger<N> &n )
{
    int   i;
    int   size = n.getSize();

    if (N < 11)
    {
        for (i = size-1; i >= 0; i--)
        {
            os << (uint) n[i];
        }// for
    }// if
    else
    {
        for (i = size-1; i >= 0; i--)
        {
            os << (uint) n[i] << ' ';
        }// for
    }// else

    return os;
}

#endif  // __TBIGINTEGER_HH
