//
//  Project  : libmisc
//  File     : TBitArray.cc
//  Author   : Ronald Kriemann
//  Purpose  : class for array holding bits
//

extern "C" {
#include <unistd.h>

#include <netinet/in.h>
}

#include "TRandom.hh"
#include "TBitArray.hh"

// init size of the used type to store bits
ulong TBitArray::sizeOfBase = sizeof(TStoreType) * 8;

//
// constructor
//
TBitArray::TBitArray ( ulong n )
{
	noOfBits = 0;
	setSize( n );
}

//
// get size and resize
//
const uint 
TBitArray::getSize () const
{
	return noOfBits;
}

void
TBitArray::setSize ( uint n )
{
	ulong  i, oldSize = noOfBits;

	noOfBits = n;

	TDynArray<TStoreType>::setSize( (noOfBits / sizeOfBase) + 
									((noOfBits % sizeOfBase) > 0 ? 1 : 0 ) );

	for (i = oldSize; i < noOfBits; i++)
		setBit( i, 0 );
}

//
// access bits (not possible via operator[])
//
const bool 
TBitArray::getBit ( uint n ) const
{
	ulong       offset;
	TStoreType  temp, mask;

	if (n < noOfBits)
	{
		temp   = (*this)[n / sizeOfBase];
		offset = n % sizeOfBase;
			   
		mask   = 1;
		mask <<= offset;

		if (temp & mask)
			return TRUE;
		else
			return FALSE;
	}// if

	return FALSE;
}

void
TBitArray::setBit ( uint n, bool b )
{
	ulong       pos, offset;
	TStoreType  temp, mask;

	if (n < noOfBits)
	{
		pos = n / sizeOfBase;

		temp   = (*this)[pos];
		offset = n % sizeOfBase;
			   
		mask   = 1;
		mask <<= offset;

		if (b)
			temp |= mask;
		else
			temp &= ~mask;

		(*this)[pos] = temp;
	}// if
}

//
// init bitarray with random bitstream
//
void
TBitArray::initRandom ( const ulong &seed )
{
	ulong  i;

	for (i = 0; i < noOfBits; i++)
	{
		if (TRandom::random( 0, 1 ) < 0.5)
			setBit( i, 0 );
		else
			setBit( i, 1 );
	}// for
}

//
// XOR operator
//
TBitArray &
TBitArray::operator ^= ( const TBitArray &a )
{
	ulong      i;
	TBitArray  temp;
	ulong      size1, size2, size;

	size1 =   getSize();
	size2 = a.getSize();

	if (size1 < size2)
	{
		temp = a;

		// XOR via CPU
		size = size1 / sizeOfBase;

		for (i = 0; i < size; i++)
		{
			temp[i] ^= (*this)[i];
		}// for

		// XOR by hand
		size = size1 % sizeOfBase;

		for (i = size1-size; i < size1; i++)
		{
			if (temp.getBit(i) == getBit(i))
				temp.setBit( i, 0 );
			else
				temp.setBit( i, 1 );
		}// for

		// all following bits are assumed to be 0
		for (i = size1; i < size2; i++)
		{
			temp.setBit( i, a.getBit(i) );
		}// for
	}// if
	else
	{
		temp = *this;

		// XOR via CPU
		size = size2 / sizeOfBase;

		for (i = 0; i < size; i++)
		{
			temp[i] ^= a[i];
		}// for

		// XOR by hand
		size = size2 % sizeOfBase;

		for (i = size2-size; i < size2; i++)
		{
			if (temp.getBit(i) == a.getBit(i))
				temp.setBit( i, 0 );
			else
				temp.setBit( i, 1 );
		}// for

		// all following bits are assumed to be 0
		for (i = size2; i < size1; i++)
		{
			temp.setBit( i, getBit(i) );
		}// for
	}// else

	return *this = temp;
}

//
// copy operator
//
TBitArray & 
TBitArray::operator = ( const TBitArray &a )
{
	TDynArray<TStoreType>::operator=( a );

	noOfBits = a.noOfBits;

	return *this;
}

TBitArray & 
TBitArray::operator = ( const TString &string )
{
	ulong       i,j,k;
	ulong       charBitSize, fac;
	ulong       strLen = string.length();
	TStoreType  temp, mask;


	charBitSize = string.charSize();

	// we assume an integral divisor
	fac = sizeOfBase / string.charSize();

	setSize( strLen * charBitSize );

	for (i = 0, k = 0; i < strLen; k++)
	{
		temp = 0;

		for (j = 0; ((j < fac) && (i < strLen)); j++, i++)
		{
			mask   = string[i];
			mask <<= j*charBitSize;

			temp |= mask;
		}// for

		(*this)[k] = temp;
	}// for

	return *this;	
}

//
// compare
//
const bool 
TBitArray::operator == ( const TBitArray &a ) const
{
	long   i;
	long   size1, size2, size;

	size1 =   getSize();
	size2 = a.getSize();

	if (size1 < size2)
	{
		// compare via CPU
		size = size1 / sizeOfBase;

		for (i = 0; i < size; i++)
		{
			if ((*this)[i] != a[i])
				return FALSE;
		}// for

		// compare by hand
		size = size1 % sizeOfBase;

		for (i = size1-size; i < size1; i++)
		{
			if (getBit(i) != a.getBit(i))
				return FALSE;
		}// for

		// all following bits are assumed to be 0
		for (i = size1; i < size2; i++)
		{
			if (a.getBit( i ) != 0)
				return FALSE;
		}// for
	}// if
	else
	{
		// compare via CPU
		size = size2 / sizeOfBase;

		for (i = 0; i < size; i++)
		{
			if ((*this)[i] != a[i])
				return FALSE;
		}// for

		// compare by hand
		size = size2 % sizeOfBase;

		for (i = size2-size; i < size2; i++)
		{
			if (getBit(i) != a.getBit(i))
				return FALSE;
		}// for

		// all following bits are assumed to be 0
		for (i = size2; i < size1; i++)
		{
			if (getBit( i ) != 0)
				return FALSE;
		}// for
	}// else

	return TRUE;
}

const bool 
TBitArray::operator != ( const TBitArray &a ) const
{
	return !((*this) == a);
}

//
// export to string
//
TString 
TBitArray::toString () const
{
	ulong    i;
	TString  temp;
	ulong    charSize = temp.charSize();
	ulong    fac = sizeOfBase / charSize;
	ulong    size, mask, stdMask;
	ulong    data1, data2;

	// set initial mask for every cycle
	stdMask = 0;
	for (i = 0; i < charSize; i += 8)
	{
		stdMask <<= 8;
		stdMask  += 0xFF;
	}// for

	size = (noOfBits / charSize) + ((noOfBits % charSize) > 0 ? 1 : 0);

	temp.setLength( size );

	for (i = 0; i < size; i++)
	{
		data1 = (*this)[i / fac];
		data2 = 0;

		mask   = stdMask;
		mask <<= (i % fac) * charSize;

		data2 = data1 & mask;
		data2 >>= (i % fac) * charSize;

		temp[i] = data2;
	}// for

	return temp;
}

//
// output
//
ostream & 
operator << ( ostream &os, const TBitArray &a )
{
	for (ulong i = 0; i < a.getSize(); i++)
	{
		if ((i > 0) && (i % 8) == 0) os << ' ';
		os << a.getBit( i );
	}// for

	return os;
}
