//
//  Project   : libmisc
//  File      : TSocket.cc
//  Author    : Ronald Kriemann
//  Purpose   : class for socket-based communication
//

extern "C" {
#include <stdio.h>
#include <unistd.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
}

#include "defines.hh"

#include "TSocket.hh"

//
// constructor
//
TSocket::TSocket ( int s )
{
	sd = s;
}

//
// destructor
//
TSocket::~TSocket ()
{
	if (sd >= 0)
		close();
}

//
// init socket descriptor
// 
int
TSocket::initSocket ()
{
	if (sd >= 0)
	{
		cerr << "TSocket::initSocket () : socket descr. already init." << endl;
		return ERROR;
	}// if

	// create a socket
	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror( "TSocket::initSocket () : socket" );
		return ERROR;
	}// if

	return SUCCESS;
}

//
// close socket
//
int 
TSocket::close ()
{
	if (sd >= 0)
	{
		if (::close( sd ) < 0)
		{
			perror( "TSocket::close () : close" );
			return ERROR;
		}// if

		sd = -1;
	}// if

	return SUCCESS;
}

//
// IO
//
int 
TSocket::read ( TString &string )
{
	char  buf[1024];
	int   n;

	if ((sd >= 0) && canRead())
	{
		if ((n = ::recv( sd, buf, sizeof( buf ), 0)) < 0)
		{
			perror( "TSocket::read (TString) : recv" );
			return ERROR;
		}// if

#ifdef DEBUG
		cout << "recv : " << buf << endl;
#endif

		string.copyFrom( buf, n );
		return SUCCESS;
	}// if

	return ERROR;
}

int 
TSocket::write ( const TString &string )
{
	char  buf[1024];
	int   n;

	if ((sd >= 0) && canWrite())
	{
		string.copyTo( buf, sizeof( buf ) );

#ifdef DEBUG
		cout << "send : " << buf << endl;
#endif

		if ((n = ::send( sd, buf, string.length(), 0)) < 0)
		{
			perror( "TSocket::write (TString) : send" );
			return ERROR;
		}// if

		return SUCCESS;
	}// if

	return ERROR;
}

//
// IO with operators
//
TSocket & 
TSocket::operator << ( const TString &string )
{
	write( string );

	return *this;
}

//
// return diff. status 
//
const bool 
TSocket::canRead () const
{
	fd_set          set;
	struct timeval  timer;

	if (sd >= 0)
	{
		FD_ZERO( &set );
		FD_SET( sd, &set );
		
		timer.tv_sec = 0;
		timer.tv_usec = 0;
		
		if (::select( sd+1, &set, 0, 0, &timer ) < 0)
		{
			perror( "TSocket::canRead () : select" );
			return FALSE;
		}// if
		else if (FD_ISSET(sd, &set)) 
			return TRUE;
	}// if

	return FALSE;
}

const bool 
TSocket::canWrite () const
{
	fd_set          set;
	struct timeval  timer;

	if (sd >= 0)
	{
		FD_ZERO( &set );
		FD_SET( sd, &set );
		
		timer.tv_sec = 0;
		timer.tv_usec = 0;
		
		if (::select( sd+1, 0, &set, 0, &timer ) < 0)
		{
			perror( "TSocket::canRead () : select" );
			return FALSE;
		}// if
		else if (FD_ISSET(sd, &set)) 
			return TRUE;
	}// if

	return FALSE;
}

//
// access port
//
const uint & 
TSocket::getPort () const
{
	return portNumber;
}

void 
TSocket::setPort ( const uint & port )
{
	portNumber = port;
}
