#ifndef __DYNARRAY_HH
#define __DYNARRAY_HH
//
//  Project  : libmisc
//  File     : dynarray.hh
//  Author   : Ronald Kriemann
//  Purpose  : class for a dyn. array of templates
//  Date     : 19.05.1997
//

#include <iostream.h>
 
#include "defines.hh"

#include "TException.hh"

//
// class for a template-vector
//
template <class T> 
class TDynArray
{
protected:
	T*    pList;        // the array with the stored elements
	uint  nSize;        // number of the element with the max. index
	uint  nRealSize;    // real size of the array
	uint  nGrowBy;      // if index > nRealSize the list will grow to index+nGrowBy+1

public:
	// constructor
	TDynArray ();
	TDynArray ( uint size, uint growby = 0 );
	TDynArray ( TDynArray<T>& vec );

	// default destructor
	virtual ~TDynArray ();

	// return data pointer
	virtual T* getData () const;

	// return size of the vector
	virtual const uint getSize () const;

	// set a new size but keep the old elements
	virtual void setSize ( uint n );

	// set the growby size new
	virtual TDynArray<T> & setGrowBy ( uint growby );

	// access the vector as an array
	virtual T& operator [] ( uint i );
	virtual const T& operator [] ( uint i ) const;

	// copy operator
	virtual TDynArray<T> & operator = ( const TDynArray<T>& vec );
};

/////////////////////////////////////////////////////
//
// implementation
//

//
// default constructor
//
template <class T> 
TDynArray<T>::TDynArray ()
{
	nSize = nRealSize = 0;
	nGrowBy = 0;
	pList = 0;
}

//
// constructor with default size
//
template <class T> 
TDynArray<T>::TDynArray ( uint size, uint growby = 0 )
{
	nSize     = size;
	nGrowBy   = growby;
	nRealSize = size+nGrowBy;
		
	pList = new T[nRealSize];
}

//
// copy constructor
//
template <class T> 
TDynArray<T>::TDynArray ( TDynArray<T>& vec ) : pList(NULL), nSize(0)
{ 
	(*this) = vec; 
}

//
// default destructor
//
template <class T> 
TDynArray<T>::~TDynArray ()
{ 
	if (pList) delete[] pList; 
}

//
// return data pointer
//
template <class T> 
T* 
TDynArray<T>::getData () const
{
	return pList;
}

//
// return size of the vector
//
template <class T> 
const uint 
TDynArray<T>::getSize () const 
{ 
	return (nSize); 
}

//
// set a new size but keep the old elements
//
template <class T> 
void
TDynArray<T>::setSize ( uint n )
{
	uint  i;

	if (nSize != n)
	{
		if (n == 0) 
		{
			if (pList) delete[] pList;
			pList = NULL;
			nSize = nRealSize = 0;
		}// if
		else if (n > nRealSize)
		{
			nRealSize = n + nGrowBy;
			T *tmpList = new T[nRealSize];
				
			if (pList != NULL)
			{
				T *p1 = tmpList;
				T *p2 = pList;
				for (i = 0; i < nSize; i++)
					*p1++ = *p2++;

				delete[] pList;
			}// if

			nSize = n;
			pList = tmpList;
		} 
		else if ((n < nSize - nGrowBy) && (nGrowBy < nSize))
		{
			nSize = n;
			nRealSize = nSize+nGrowBy;
				
			T *tmpList = new T[nRealSize];

			if (pList != NULL)
			{
				T *p1 = tmpList;
				T *p2 = pList;
				for (i = 0; i < n; i++)
					*p1++ = *p2++;
					
				delete[] pList;
			}// if
				
			pList = tmpList;
		}
		else // if (n < nSize)
		{
			nSize = n;
		} 
	}// if
}

//
// set the growby size new
//
template <class T> 
TDynArray<T> &
TDynArray<T>::setGrowBy ( uint growby ) 
{
	nGrowBy = growby;

    return *this;
}

//
// access the vector as an array
//
template <class T> 
T& 
TDynArray<T>::operator [] ( uint i ) 
{ 
	if (i >= nSize)
    {
#if defined(EXCPETIONS)
        throw TRange( i, nSize );
#else
        cerr << "TDynArray::operator [] : range overflow (pos = " << i
             << ", max = " << nSize << ")" << endl;
#endif
    }// if
    
	return *(pList+i); 
}

//
// access the vector as an array (const access)
//
template <class T> 
const T& 
TDynArray<T>::operator [] ( uint i ) const 
{ 
	if (i >= nSize)
    {
#if defined(EXCPETIONS)
        throw TRange( i, nSize );
#else
        cerr << "TDynArray::operator [] : range overflow (pos = " << i
             << ", max = " << nSize << ")" << endl;
#endif
    }

	return *(pList+i); 
}

//
// copy operator
//
template <class T> 
TDynArray<T>& 
TDynArray<T>::operator = ( const TDynArray<T>& vec ) 
{
	uint  i;

	if (vec.nSize == nSize)
	{
		nGrowBy = vec.nGrowBy;

		T *p1 = pList;
		T *p2 = vec.pList;

		for (i = 0; i < nSize; i++)
			*p1++ = *p2++;
	}// if
	else 
	{
		nGrowBy   = vec.nGrowBy;
		nSize     = vec.nSize;
		nRealSize = vec.nRealSize;

		// set new size but no need for keeping old elements
		if (pList != NULL) delete[] pList;
		pList = new T[nRealSize];

		if (nSize > 0)
		{
			T *p1 = pList;
			T *p2 = vec.pList;
			for (i = 0; i < nSize; i++)
				*p1++ = *p2++;
		}// if
	}// else
	return (*this);
}

#endif			// __DYNARRAY_HH
