//
//  Project   : libmisc
//  File      : idea.cc
//  Author    : Christian Rost
//  Purpose   : class for Idea encoding and decoding
//

extern "C"
{
#include <string.h>
}

#include "idea.hh"

#define LSW16(y) ((y) & 0xffff)                  /* low significant 16-bit */
#define MSW16(y) ((y >> 16) & 0xffff)            /* most significant 16-bit */
#define LSW8(y) ((y) & 0xff)                     /* low significant 8-bit */
#define MSW8(y) ((y >> 8) & 0xff)                /* most significant 8-bit */
#define MUL_MOD  (ulong)(((ulong)1 << 16) | 1)   /* 2**16 + 1 */

TIdea::TIdea () {modus=0;}
TIdea::~TIdea () {}

//Multiplication modulo 2**16 + 1
ushort TIdea::mul( ushort x , ushort y)
{
    ulong q;
    long p;
    if (!x) 
        return LSW16(MUL_MOD - (ulong)y);
    else 
        if (!y) 
            return LSW16(MUL_MOD - (ulong)x);
        else 
        {
            q = (ulong)x * (ulong)y;
            p = LSW16(q) - MSW16(q);
            if (p <= 0) p += MUL_MOD;
            return LSW16(p);
        }
} 

// Compute multiplicative inverse of x by Euclid's GCD algorithm
ushort TIdea::mul_inv(ushort x)
{
    long n1 = MUL_MOD;
    long n2 = (ulong)x;
    long b1 = 0;
    long b2 = 1;
    long q, r, t;
    if (x <= 1) return x;
    while (1) {
        r = n1 % n2;
        q = n1 / n2;
        if (!r) {
            if (b2 < 0) b2 += MUL_MOD;
            return LSW16(b2);
        }
        else {
            n1 = n2;
            n2 = r;
            t  = b2;
            b2 = b1 - q * b2;
            b1 = t;
        }
    }
}

//create the set of subkeys for encryption
void TIdea::encrypt_subkeys(idea_key_t key, 
                            idea_subkeys_t subkeys)
{
    int i;  
    //setting main-key
    for (i=0;i<8;i++)
        main_key[i]=key[i];
  
  
    for (i=0;i<8;i++) subkeys[i] = key[i];
    for (;i<IDEA_SK_NUM;i++)
        subkeys[i] = LSW16((subkeys[(i+1) % 0x8 ? i-7 : i-15] << 9) |
                           (subkeys[(i+2) % 0x8 < 2 ? i-14 : i-6] >> 7));
}

//create the set of subkeys for decryption  
void TIdea::decrypt_subkeys( idea_subkeys_t encrypt_subkeys, 
                             idea_subkeys_t decrypt_subkeys)
{
    ushort *pen = (ushort*)encrypt_subkeys;
    ushort *pde = (ushort*)decrypt_subkeys;
    idea_subkeys_t t;
    ushort *pt = (ushort*)t;
    int i;

    t[6*IDEA_ROUNDS] = mul_inv(*pen++);
    t[6*IDEA_ROUNDS+1] = LSW16(-*pen++);
    t[6*IDEA_ROUNDS+2] = LSW16(-*pen++);
    t[6*IDEA_ROUNDS+3] = mul_inv(*pen++);
    for (i=6*(IDEA_ROUNDS-1);i>=0;i-=6) 
    {
        t[i + 4] = *pen++;
        t[i + 5] = *pen++;
        t[i] = mul_inv(*pen++);
        if (i) 
        {
            t[i + 2] = LSW16(-*pen++);
            t[i + 1] = LSW16(-*pen++);
        }
        else {
            t[1] = LSW16(-*pen++);
            t[2] = LSW16(-*pen++);
        }
        t[i + 3] = mul_inv(*pen++);
    }
    for (i=0;i<IDEA_SK_NUM;i++) 
    {
        *pde++ = *pt;
        *pt++ = 0;
    }
}

//Idea Block cipher (encryption AND decryption)
void TIdea::cipher(idea_block_t inblock,
                   idea_block_t outblock,
                   idea_subkeys_t key)
  
{
    ushort *pin = (ushort*)inblock;
    ushort *pout = (ushort*)outblock;
    ushort *pk = (ushort*)key;
    ushort word1, word2, word3, word4;
    ushort t1, t2;
    int i;

    word1 = *pin++;
    word2 = *pin++;
    word3 = *pin++;
    word4 = *pin;

    for (i=IDEA_ROUNDS;i>0;i--) 
    {
        word1 = mul(word1,*pk++);
        word2 = LSW16(word2 + *pk++);
        word3 = LSW16(word3 + *pk++);
        word4 = mul(word4,*pk++);

        t2 = word1 ^ word3;
        t2 = mul(t2,*pk++);
        t1 = LSW16(t2 + (word2 ^ word4));
        t1 = mul(t1,*pk++);

        word1 ^= t1;
        word4 ^= t2;

        t2 ^= word2;
        word2 = word3 ^ t1;
        word3 = t2;
    }
 
    word1 = mul(word1,*pk++);
    *pout++ = word1;
    *pout++ = LSW16(word3 + *pk++);
    *pout++ = LSW16(word2 + *pk++);
    word4 = mul(word4,*pk);
    *pout = word4;
}

//encrypt creates the encryption key-set with an md5(text)-key
void 
TIdea::buildEncryptKeys( const TString & text )
{
    TString    digest;
    idea_key_t ideakey;
    TMD5       md5;
    
    digest = md5.crypt( text );

    for (int i = 0; i < 8; i++) 
    {
        ideakey[i]   = 0;
        ideakey[i]   = digest[2*i];
        ideakey[i] <<= 8;
        ideakey[i]  |= digest[(2*i)+1];     
    }// for

    encrypt_subkeys(ideakey,enc_subkeys);
}

//decrypt creates the decryption key-set with an md5(text)-key
void 
TIdea::buildDecryptKeys( const TString & text )
{
    TString        digest;
    idea_key_t     ideakey;
    idea_subkeys_t buffer;
    TMD5           md5;
    
    digest = md5.crypt( text );
    
    for (int i=0; i<8; i++) 
    {
        ideakey[i]   = 0;
        ideakey[i]   = digest[2*i];
        ideakey[i] <<= 8;
        ideakey[i]  |= digest[(2*i)+1];     
    }
    
    encrypt_subkeys( ideakey, buffer );
    decrypt_subkeys( buffer, de_subkeys );
}
  
void
TIdea::generateSubKeys(const TString & text)
{
    if (text == "")
    {
        // use the imported key
        encrypt_subkeys( main_key   , enc_subkeys );
        decrypt_subkeys( enc_subkeys, de_subkeys );
    }// if
    else 
    {
        // make a new key from md5(text)
        buildEncryptKeys( text );
        buildDecryptKeys( text );
    }// else
}
  

void
TIdea::cypher_textblock(uchar out[8], uchar in[8] )
{
    idea_block_t block;
    for (int i=0; i<4; i++) 
    {
        block[i]=0; block[i]=in[2*i];
        block[i] <<= 8; block[i] |= in[(2*i)+1];     
    }
   
    if (modus)
        //decryption
        cipher(block,block,de_subkeys);
    else
        //encryption
        cipher(block,block,enc_subkeys);

    for (int i=0; i<4; i++) 
    {
        out[2*i]=MSW8(block[i]); out[(2*i)+1]=LSW8(block[i]);    
    }  
}


// VORSICHT: out muss aureicheichend gross allokiert werden !
// Aufgerundet auf eine Vielfaches von 8 bytes +1 (Terminierende 0) !!!
void                                         
TIdea::cypher_text(int mod, char* out, const char* in, int inlen=0)
{
    uchar puffer[8];
    uchar cryptOut[8];

    // Falls in ein "echter" string (zum beispiel beim encoden)
    // kann die laenge von in ueber strlen ermittelt werden
    // Beim decoden ist in im allgemeinen kein "echter" string, die
    // Laenge muss also explizit uebergeben werden  
    modus=mod;
    int size;
    if (modus==0)
        if (inlen==0)
        {
            size=(strlen(in)/8)*8+1;
            if (strlen(in)%8>0) size+=8;      
            inlen=strlen(in);
        }
        else 
        {      
            size=(inlen/8)*8+1;
            if (inlen%8>0) size+=8;        
        }  
    else size=inlen;
  
    char tmpout[size];
    memset(tmpout, 0, sizeof(tmpout) ); 
    int start=0;
    for (int i=0; i<inlen; i++)
    {
        if (i%8==0) memset(puffer, 0, sizeof(puffer) );
        puffer[i%8]=in[i];
        if (i%8==7 || i==inlen-1)
        {
            cypher_textblock(cryptOut,puffer);
            for (int ii=0;ii<8;ii++)
            {	      
                tmpout[start]=cryptOut[ii];
                //for (int iii=0; iii<size;iii++) printf("%X ", tmpout[iii]);printf("\n");
                start++;			     
            }
        }    	
    }  
    for (int i=0; i<size;i++)    
        out[i]=tmpout[i];
    out[size-1]=0;
}

void
TIdea::encode(char* out, const char* in, int length)
{
    cypher_text(0,out,in,length);
}

void
TIdea::decode(char* out, const char* in,int inlen)
{
    cypher_text(1,out,in,inlen);
}

TString
TIdea::encode(const TString &in)
{
    char tmp[in.length()+1];
    in.copyTo(tmp,in.length()+1);  

    TString result;
    int size=(in.length()/8)*8+1;
    if (in.length()%8>0) size+=8;   
    char tmpresult[size];
    encode(tmpresult,tmp,in.length());
    result.copyFrom(tmpresult,size); 
    return result;
}

TString
TIdea::decode(const TString &in)
{
    TString result; 
    char tmp[in.length()+1];
    char tmpresult[in.length()+1];
    in.copyTo(tmp,in.length());
    decode(tmpresult,tmp,in.length());
    result.copyFrom(tmpresult);
    return result;
}




void
TIdea::exportKey(idea_key_t out)
{
    for (int i=0; i<8; i++)
        out[i]=main_key[i];
}

void
TIdea::importKey(idea_key_t in)
{
    for (int i=0; i<8; i++)
        main_key[i]=in[i];
}

//delete subkey-set from memory
void
TIdea::clearkey()  
{
    for (int i=0; i<IDEA_SK_NUM; i++)
    {
        enc_subkeys[i]=0;
        de_subkeys[i]=0;
    }
}
